import numpy as np
from dataclasses import dataclass
from typing import List, Generic

from .common import Result
from ..models import _exponential_model
from ..types import FloatLike, F


@dataclass(eq=True, frozen=True)
class ExponentialDecayMode(Generic[F]):
    amplitude: F
    decay_rate: float
    offset: F

    def __call__(self, t: FloatLike) -> FloatLike:
        return _exponential_model(
            t, 0.0, np.array([self.amplitude]), np.array([self.decay_rate])
        )


@dataclass
class ExponentialDecayResult(Result, Generic[F]):
    amplitudes: np.ndarray[F]
    decay_rates: np.ndarray[float]

    def __post_init__(self):
        order = np.argsort(self.decay_rates)
        self.amplitudes = self.amplitudes[order]
        self.decay_rates = self.decay_rates[order]

    @property
    def modes(self) -> List[ExponentialDecayMode]:
        return [
            ExponentialDecayMode(amplitude=amplitude, decay_rate=decay_rate, offset=self.offset)
            for amplitude, decay_rate in zip(self.amplitudes, self.decay_rates)
        ]

    def __call__(self, t: FloatLike) -> np.ndarray[FloatLike]:
        return _exponential_model(t, self.offset, self.amplitudes, self.decay_rates)

    def __repr__(self):
        return f"ExponentialDecayResult(offset={self.offset}, modes={self.modes})"

    def pretty_repr(self):
        if np.iscomplex(self.offset):
            offset_str = f"offset = {self.offset.real:0.2f} + {self.offset.imag:0.2f}j"
        else:
            offset_str = f"offset = {self.offset:0.2f}"


        if len(self.amplitudes) == 1:
            if np.iscomplex(self.amplitudes[0]):
                amplitude_str = f"{self.amplitudes[0].real:0.2e} + {self.amplitudes[0].imag:0.2e}j"
            else:
                amplitude_str = f"{self.amplitudes[0]:0.2e}"

            return f"offset = {offset_str}, amplitude = {amplitude_str}, decay_rate / 2pi = {self.decay_rates[0] / (2*np.pi):0.2e}"
        else:
            return f"offset = {offset_str}, {len(self.amplitudes)} modes"



from .complex_mode import ComplexMode, ComplexResult
from .damped_cosine_mode import DampedCosineMode, DampedCosineResult
from .exponential_decay_mode import ExponentialDecayResult, ExponentialDecayMode

__all__ = [
    "ComplexMode",
    "ComplexResult",
    "DampedCosineMode",
    "DampedCosineResult",
    "ExponentialDecayResult",
    "ExponentialDecayMode",
]



import numpy as np
from dataclasses import dataclass
from typing import List

from .common import Mode, Result
from ..models import _complex_exponential_model
from ..types import FloatLike


@dataclass(eq=True, frozen=True)
class ComplexMode(Mode):
    complex_amplitude: complex

    @property
    def amplitude(self):
        return np.abs(self.complex_amplitude)

    @property
    def phase(self):
        return np.angle(self.complex_amplitude)

    def __call__(self, t: FloatLike) -> FloatLike:
        return _complex_exponential_model(
            t,
            0,
            np.array([self.complex_amplitude]),
            np.array([self.pulsation]),
            np.array([self.decay_rate]),
        )


@dataclass
class ComplexResult(Result):
    amplitudes: np.ndarray[complex]
    pulsations: np.ndarray[float]
    decay_rates: np.ndarray[float]

    def __post_init__(self):
        order = np.argsort(self.decay_rates)
        self.amplitudes = self.amplitudes[order]
        self.pulsations = self.pulsations[order]
        self.decay_rates = self.decay_rates[order]

    @property
    def modes(self) -> List[ComplexMode]:
        return [
            ComplexMode(complex_amplitude=amplitude, pulsation=pulsation, decay_rate=decay_rate)
            for amplitude, pulsation, decay_rate in zip(self.amplitudes, self.pulsations, self.decay_rates)
        ]

    @property
    def frequencies(self) -> np.ndarray[float]:
        return self.pulsations / (2 * np.pi)

    def __call__(self, t: FloatLike) -> np.ndarray[complex]:
        return _complex_exponential_model(
            t, self.offset, self.amplitudes, self.pulsations, self.decay_rates
        )

    def __repr__(self):
        return f"ComplexResult(offset={self.offset}, modes={self.modes})"

    def pretty_repr(self):
        if len(self.amplitudes) == 1:
            return f"offset = {self.offset.real:0.2f} + {self.offset.imag:0.2f}j, amplitude = {self.amplitudes[0].real:0.2e}+{self.amplitudes[0].imag:0.2e}j,  pulsation = {self.pulsations[0]:0.2e}, decay_rate / 2pi = {self.decay_rates[0] / (2*np.pi):0.2e}"
        else:
            return f"offset = {self.offset.real:0.2f} + {self.offset.imag:0.2f}j, {len(self.amplitudes)} modes"




from typing import Generic

import numpy as np
from dataclasses import dataclass

from ..types import FloatLike, F


@dataclass(eq=True, frozen=True)
class Mode:
    pulsation: float  # omega (pulsation)
    decay_rate: float

    @property
    def frequency(self):
        return 2 * np.pi * self.pulsation

    def __call__(self, t: FloatLike) -> FloatLike:
        raise NotImplementedError()


@dataclass
class Result(Generic[F]):
    offset: F

    times: np.ndarray[float]
    signal: np.ndarray[F]

    def plot(self):
        from ..plot import plot

        plot(self)



import numpy as np
from dataclasses import dataclass

from .common import Mode, Result
from ..models import _damped_cosine_model
from ..types import FloatLike


@dataclass(eq=True, frozen=True)
class DampedCosineMode(Mode):
    amplitude: float
    phase: float

    def __call__(self, t: FloatLike) -> FloatLike:
        return _damped_cosine_model(
            t,
            0.0,
            np.array([self.amplitude]),
            np.array([self.phase]),
            np.array([self.pulsation]),
            np.array([self.decay_rate]),
        )


@dataclass
class DampedCosineResult(Result):
    offset: float
    amplitudes: np.ndarray[float]
    phases: np.ndarray[float]
    pulsations: np.ndarray[float]
    decay_rates: np.ndarray[float]

    @property
    def modes(self) -> list[DampedCosineMode]:
        return [
            DampedCosineMode(amplitude=amplitude, phase=phase, pulsation=pulsation, decay_rate=decay_rate)
            for amplitude, phase, pulsation, decay_rate in zip(
                self.amplitudes, self.phases, self.pulsations, self.decay_rates
            )
        ]

    @property
    def frequencies(self) -> np.ndarray[float]:
        return self.pulsations / (2 * np.pi)

    def __call__(self, t: FloatLike) -> FloatLike:
        return _damped_cosine_model(
            t, self.offset, self.amplitudes, self.phases, self.pulsations, self.decay_rates
        )

    def __repr__(self):
        return f"DampedCosineResult(offset={self.offset}, modes={self.modes})"

    def pretty_repr(self):
        if len(self.amplitudes) == 1:
            return f"offset = {self.offset:0.2f}, amplitude = {self.amplitudes[0]:0.2e}, phase = {self.phases[0]:0.2f}, pulsation = {self.pulsations[0]:0.2e}, decay_rate / 2pi = {self.decay_rates[0] / (2*np.pi):0.2e}"
        else:
            return f"offset = {self.offset:0.2f}, {len(self.amplitudes)} modes"



import functools as ft
from dataclasses import dataclass
from typing import Callable, Tuple

import numpy as np
from scipy.optimize import minimize

from .models import _damped_cosine_model, _complex_exponential_model, _exponential_model
from .results import DampedCosineResult
from .types import FloatLike

NO_BOUND = (None, None)
POSITIVE_BOUND = (0, None)
ZERO_BOUND = (0, 0)

# =====================================================================================================================
# Post fit options
# =====================================================================================================================

@dataclass
class NoOffset:
    pass

_PostFitOptions = None | NoOffset

# =====================================================================================================================
# Common functions
# =====================================================================================================================

def _cost(
    x: np.ndarray,
    times: np.ndarray,
    signal: np.ndarray,
    model: Callable[[np.ndarray, np.ndarray], np.ndarray],
) -> float:
    model_signal = model(times, x)
    return np.sum(np.abs(model_signal - signal) ** 2)


# =====================================================================================================================
# Complex exponential model
# =====================================================================================================================


def _complex_exponential_adapter(t: np.ndarray[float], x: np.ndarray[FloatLike]):
    offset_re, offset_im, modes = x[0], x[1], x[2:].reshape(-1, 4)
    offset = offset_re + 1j * offset_im

    amplitudes_re, amplitudes_im, pulsations, decay_rates = (
        modes[:, 0],
        modes[:, 1],
        modes[:, 2],
        modes[:, 3],
    )
    amplitudes = amplitudes_re + 1j * amplitudes_im
    return _complex_exponential_model(t, offset, amplitudes, pulsations, decay_rates)


def _post_fit_complex_exponential(
    times: np.ndarray,
    signal: np.ndarray,
    offset: complex,
    amplitudes: np.ndarray[complex],
    pulsations: np.ndarray[float],
    decay_rates: np.ndarray[float],
    options: _PostFitOptions
) -> Tuple[complex, np.ndarray[complex], np.ndarray[float], np.ndarray[float]]:
    assert len(amplitudes) == len(pulsations) == len(decay_rates)

    cost = ft.partial(
        _cost, times=times, signal=signal, model=_complex_exponential_adapter
    )

    x0 = [offset.real, offset.imag]
    x1 = np.stack((amplitudes.real, amplitudes.imag, pulsations, decay_rates)).T.flatten()
    x0 = np.concatenate((x0, x1))

    if isinstance(options, NoOffset):
        offset_bounds = [ZERO_BOUND, ZERO_BOUND]  # Force offset to zero
    else:
        offset_bounds = [NO_BOUND, NO_BOUND]  # No bounds for offset
    bounds = offset_bounds
    bounds += [NO_BOUND, NO_BOUND, NO_BOUND, POSITIVE_BOUND] * len(amplitudes)

    xopt = minimize(cost, x0, bounds=bounds).x
    offset = xopt[0] + 1j * xopt[1]
    amplitudes_re, amplitudes_im, pulsations, decay_rates = xopt[2:].reshape(-1, 4).T
    amplitudes = amplitudes_re + 1j * amplitudes_im

    return offset, amplitudes, pulsations, decay_rates


# =====================================================================================================================
# Real exponential model
# =====================================================================================================================


def _exponential_adapter(
    t: np.ndarray[float], x: np.ndarray[FloatLike], is_complex: bool
) -> np.ndarray[FloatLike]:
    if is_complex:
        offset = complex(x[0] + 1j * x[1])
        modes = x[2:].reshape(-1, 3)
        amplitudes_re, amplitudes_im, decay_rates = modes[:, 0], modes[:, 1], modes[:, 2]
        amplitudes = amplitudes_re + 1j * amplitudes_im
    else:
        offset = complex(x[0])
        modes = x[1:].reshape(-1, 2)
        amplitudes, decay_rates = modes[:, 0], modes[:, 1]

    return _exponential_model(t, offset, amplitudes, decay_rates)


def _post_fit_exponential(
    times: np.ndarray,
    signal: np.ndarray,
    offset: complex,
    amplitudes: np.ndarray[complex],
    decay_rates: np.ndarray[complex],
    is_complex: bool,
    options: _PostFitOptions
) -> Tuple[FloatLike, np.ndarray[FloatLike], np.ndarray[FloatLike]]:
    assert len(amplitudes) == len(decay_rates)

    cost = ft.partial(
        _cost,
        times=times,
        signal=signal,
        model=ft.partial(_exponential_adapter, is_complex=is_complex),
    )

    if isinstance(options, NoOffset):
        if is_complex:
            offset_bounds = [ZERO_BOUND, ZERO_BOUND]  # Force offset to zero
            offset_guess = [0, 0]
        else:
            offset_bounds = [ZERO_BOUND]  # Force offset to zero
            offset_guess = [0]
    else:
        if is_complex:
            offset_bounds = [NO_BOUND, NO_BOUND]  # No bounds for offset
            offset_guess = [offset.real, offset.imag]
        else:
            offset_bounds = [NO_BOUND]  # No bounds for offset
            offset_guess = [offset.real]

    if is_complex:
        x0 = np.stack((amplitudes.real, amplitudes.imag, decay_rates)).T.flatten()
        x0 = np.concatenate((offset_guess, x0))

        bounds = offset_bounds + [NO_BOUND, NO_BOUND, POSITIVE_BOUND] * len(
            amplitudes
        )
    else:
        x0 = np.stack((amplitudes.real, decay_rates)).T.flatten()
        x0 = np.concatenate((offset_guess, x0))
        bounds = offset_bounds + [NO_BOUND, POSITIVE_BOUND] * len(amplitudes)

    x0 = np.array(x0)

    xopt = minimize(cost, x0, bounds=bounds).x
    if is_complex:
        offset = xopt[0] + 1j * xopt[1]
        amplitudes_re, amplitudes_im, decay_rates = xopt[2:].reshape(-1, 3).T
        amplitudes = amplitudes_re + 1j * amplitudes_im
    else:
        offset = xopt[0]
        amplitudes, decay_rates = xopt[1:].reshape(-1, 2).T

    return offset, amplitudes, decay_rates


# =====================================================================================================================
# Damped cosine model
# =====================================================================================================================


def _damped_cosine_adapter(
    t: np.ndarray[float], x: np.ndarray[float]
) -> np.ndarray[float]:
    offset, modes = float(x[0]), x[1:].reshape(-1, 4)
    amplitudes, phases, pulsations, decay_rates = modes[:, 0], modes[:, 1], modes[:, 2], modes[:, 3]
    return _damped_cosine_model(t, offset, amplitudes, phases, pulsations, decay_rates)


def _post_fit_damped_cosine(
    times: np.ndarray[float],
    signal: np.ndarray[float],
    offset: complex,
    amplitudes: np.ndarray[float],
    phases: np.ndarray[float],
    pulsations: np.ndarray[float],
    decay_rates: np.ndarray[float],
    options: _PostFitOptions
) -> DampedCosineResult:
    cost = ft.partial(_cost, times=times, signal=signal, model=_damped_cosine_adapter)

    if isinstance(options, NoOffset):
        offset_guess = [0.0]
        offset_bounds = ZERO_BOUND
    else:
        offset_guess = [offset]
        offset_bounds = NO_BOUND

    x0 = np.stack((amplitudes, phases, pulsations, decay_rates)).T.flatten()
    x0 = np.concatenate((offset_guess, x0))

    bounds = [offset_bounds] + [NO_BOUND, NO_BOUND, POSITIVE_BOUND, POSITIVE_BOUND] * len(amplitudes)
    xopt = minimize(cost, x0, bounds=bounds).x
    offset = xopt[0]

    amplitudes, phases, pulsations, decay_rates = xopt[1:].reshape(-1, 4).T

    new_result = DampedCosineResult(
        times=times,
        signal=signal,
        offset=offset,
        amplitudes=amplitudes,
        pulsations=pulsations,
        decay_rates=decay_rates,
        phases=phases,
    )

    return new_result



from typing import List

import matplotlib.pyplot as plt
import numpy as np

from matplotlib.axes._axes import Axes

from .results import ComplexResult, ExponentialDecayResult, DampedCosineResult
from .types import FloatLike


def plot(
    result: ComplexResult | DampedCosineResult | ExponentialDecayResult[FloatLike],
    axs: Axes | List[Axes] | None = None,
):
    if np.iscomplexobj(result.signal):
        if axs is not None:
            if not isinstance(axs, (list, np.ndarray)):
                raise TypeError(
                    "Expected axs to be a list or numpy array of matplotlib.axes.Axes or None"
                )
            elif len(axs) != 3:
                raise ValueError(
                    "Expected axs to contain exactly 3 Axes for complex plot"
                )

        _plot_complex(result, axs)  # ty: ignore[invalid-argument-type]
    else:
        if axs is not None and not isinstance(axs, Axes):
            raise TypeError("Expected axs to be a matplotlib.axes.Axes or None")
        _plot_real(result, axs)  # ty: ignore[invalid-argument-type]


def _plot_complex(
    result: ComplexResult | ExponentialDecayResult[complex], axs: List[Axes] | None
):
    fig = None
    if axs is None:
        fig, axs = plt.subplots(1, 3, figsize=(15, 5))

    if fig is not None:
        fig.suptitle("Fit result: " + result.pretty_repr())

    dense_times = np.linspace(result.times[0], result.times[-1], 1000)
    axs[0].plot(result.times, result.signal.real, label="Real Part")
    axs[0].plot(
        dense_times, result(dense_times).real, label="Fitted Real Part", linestyle="--"
    )
    axs[0].set_xlabel("Time")
    axs[0].set_ylabel("Real Amplitude")
    axs[0].legend()

    axs[1].plot(result.times, result.signal.imag, label="Imaginary Part")
    axs[1].plot(
        dense_times,
        result(dense_times).imag,
        label="Fitted Imaginary Part",
        linestyle="--",
    )
    axs[1].set_xlabel("Time")
    axs[1].set_ylabel("Imaginary Amplitude")
    axs[1].legend()

    axs[2].plot(result.signal.real, result.signal.imag, label="Signal")
    axs[2].plot(
        result(dense_times).real,
        result(dense_times).imag,
        label="Fitted Signal",
        linestyle="--",
    )
    axs[2].set_xlabel("Real Amplitude")
    axs[2].set_ylabel("Imaginary Amplitude")
    axs[2].legend()


def _plot_real(result: DampedCosineResult | ExponentialDecayResult[float], ax: Axes | None):
    fig = None
    if ax is None:
        fig, ax = plt.subplots()

    if fig is not None:
        fig.suptitle("Fit result: " + result.pretty_repr())

    dense_times = np.linspace(result.times[0], result.times[-1], 1000)
    ax.plot(result.times, result.signal, label="Signal")
    ax.plot(dense_times, result(dense_times), label="Fitted Signal", linestyle="--")
    ax.set_xlabel("Time")
    ax.set_ylabel("Signal Amplitude")
    ax.legend()



from typing import Tuple

import numpy as np
from scipy.optimize import linear_sum_assignment

from .post_fitting import (
    _post_fit_complex_exponential,
    _post_fit_damped_cosine,
    _post_fit_exponential,
    NoOffset,
)
from .results import ComplexResult, ExponentialDecayResult, DampedCosineResult

PostFitOptions = bool | NoOffset

def fit_complex_exponential(
    times: np.ndarray[float],
    signal: np.ndarray[complex],
    n_modes: int = 1,
    post_fit: PostFitOptions = True,
    tol: float = 1e-3,
    L_fraction: float = 0.3,
) -> ComplexResult:
    r"""
    Fits a complex exponential signal of the form
    $f(t) = x_0 + \sum_k A_k \exp((j\omega_k - \kappa_k) t)$

    :param times: np.ndarray[float]
        The time points at which the signal is sampled.
    :param signal: np.ndarray[complex]
        The complex signal to fit.
    :param n_modes: int
        The number of modes to fit. Each mode corresponds to a complex exponential term.
    :param post_fit: PostFitOptions (bool | NoOffset)
        Whether to perform a post-fit refinement of the parameters, using least-square error minimization.
        This is recommended to improve the accuracy of the fit. If `False`, no post-fit is performed.
        If `True`, a full post-fit is performed. If `NoOffset()`, the post-fit is performed with the constraint that
        the offset is zero.
    :param tol: float
        The tolerance for the fitting process. It is used to determine the convergence of the fit.
    :param L_fraction: float
        The fraction of the signal length to use for the pencil method. Usually a value between 1/3 and 1/2 is
        recommended
    :return:
    """


    offset, amplitudes, pulsations, decay_rates = bicfit(
        times, signal, n_modes=n_modes, tol=tol, L_fraction=L_fraction
    )
    if post_fit:
        options = None if post_fit is True else post_fit
        offset, amplitudes, pulsations, decay_rates = _post_fit_complex_exponential(
            times, signal, offset, amplitudes, pulsations, decay_rates, options
        )

    return ComplexResult(
        times=times,
        signal=signal,
        offset=offset,
        amplitudes=amplitudes,
        pulsations=pulsations,
        decay_rates=decay_rates,
    )


def fit_exponential_decay(
    times: np.ndarray[float],
    signal: np.ndarray[float | complex],
    n_modes: int = 1,
    post_fit: PostFitOptions = True,
    is_complex: bool = False,
    tol: float = 1e-3,
    L_fraction: float = 0.3,
) -> ExponentialDecayResult:
    r""" Fits an exponential decay signal of the form
    $f(t) = x_0 + \sum_k A_k \exp(-\kappa_k t)$

    :param times: np.ndarray[float]
        The time points at which the signal is sampled.
    :param signal: np.ndarray[float | complex]
        The signal to fit. It can be real or complex.
    :param n_modes: int
        The number of modes to fit. Each mode corresponds to an exponential decay term.
    :param post_fit: PostFitOptions (bool | NoOffset)
        Whether to perform a post-fit refinement of the parameters, using least-square error minimization.
        This is recommended to improve the accuracy of the fit. If `False`, no post-fit is performed.
        If `True`, a full post-fit is performed. If `NoOffset()`, the post-fit is performed with the constraint that
        the offset is zero.
    :param is_complex: bool
        Whether the signal is complex. If True, the with will take the $(A_k)$ complex, and real otherwise.
    :param tol: float
        The tolerance for the fitting process. It is used to determine the convergence of the fit.
    :param L_fraction: float
        The fraction of the signal length to use for the pencil method. Usually a value between 1/3 and 1/2 is
        recommended.
    """
    offset, amplitudes, pulsations, decay_rates = bicfit(
        times,
        signal,
        n_modes=n_modes,
        tol=tol,
        L_fraction=L_fraction,
    )
    if post_fit:
        options = None if post_fit is True else post_fit
        offset, amplitudes, decay_rates = _post_fit_exponential(
            times, signal, offset, amplitudes, decay_rates, is_complex, options
        )
    else:
        if not is_complex:
            amplitudes = amplitudes.real
            offset = offset.real

    result = ExponentialDecayResult(
        times=times,
        signal=signal,
        offset=offset,
        amplitudes=amplitudes,
        decay_rates=decay_rates,
    )

    return result


def fit_damped_cosine(
    times: np.ndarray[float],
    signal: np.ndarray[float],
    n_modes: int = 1,
    post_fit: PostFitOptions = True,
    tol: float = 1e-3,
    L_fraction: float = 0.3,
) -> DampedCosineResult:
    r"""
    Fits a damped cosine signal of the form
    $f(t) = x_0 + \sum_k A_k \exp(-\kappa_k t) \cos(\omega_k t + \phi_k)$

    :param times: np.ndarray[float]
        The time points at which the signal is sampled.
    :param signal: np.ndarray[float]
        The real signal to fit.
    :param n_modes: int
        The number of modes to fit. Each mode corresponds to a damped cosine term.
    :param post_fit: PostFitOptions (bool | NoOffset)
        Whether to perform a post-fit refinement of the parameters, using least-square error minimization.
        This is recommended to improve the accuracy of the fit. If `False`, no post-fit is performed.
        If `True`, a full post-fit is performed. If `NoOffset()`, the post-fit is performed with the constraint that
        the offset is zero.
    :param tol: float
        The tolerance for the fitting process. It is used to determine the convergence of the fit.
    :param L_fraction: float
        The fraction of the signal length to use for the pencil method. Usually a value between 1/3 and 1/2 is
        recommended.
    :return:
    """
    # since the signal is real, there are two exponential per term
    # since 2cos(x) = exp(ix) + exp(-ix)

    offset, amplitudes, pulsations, decay_rates = bicfit(
        times,
        signal,
        n_modes=2 * n_modes,
        tol=tol,
        L_fraction=L_fraction,
    )

    amplitudes, phases, pulsations, decay_rates = _match_real_modes(amplitudes, pulsations, decay_rates, tol=tol)
    if abs(offset.imag) > tol:
        raise RuntimeError(
            f"Expected the offset to be real, but got {offset.imag} imaginary part, above fixed tolerance {tol}"
        )
    offset = offset.real

    if post_fit:
        options = None if post_fit is True else post_fit
        result = _post_fit_damped_cosine(
            times, signal, offset, amplitudes, phases, pulsations, decay_rates, options
        )
    else:
        result = DampedCosineResult(
            times=times,
            signal=signal,
            offset=offset,
            amplitudes=amplitudes,
            phases=phases,
            pulsations=pulsations,
            decay_rates=decay_rates,
        )

    return result


def bicfit(
    times: np.ndarray[float],
    signal: np.ndarray[complex],
    n_modes: int = 1,
    tol: float = 1e-3,
    L_fraction: float = 0.3,
) -> Tuple[complex, np.ndarray[complex], np.ndarray[float], np.ndarray[float]]:
    """
    Fits a signal of the form s(t) = sum_k a_k exp(x_k t)
    using a pencil method. This method is exposed for advanced users
    who want to use it directly, but it is recommended to use the
    `fit_complex_exponential`, `fit_exponential_decay` or
    `fit_damped_cosine` functions instead.

    The algorithm is taken from
    **Generalized Pencil-of-Function Method for Extracting Poles
    of an EM System from Its Transient Response**
    from Hua and Sarkar (IEEE TRANSACTIONS ON ANTENNAS
    AND PROPAGATION, VOL. 37, NO. 2, FEBRUARY 1989)
    """

    if n_modes < 1:
        raise ValueError(f"Expected at least one mode to find, got {n_modes}")

    if times.shape != signal.shape or times.ndim != 1:
        raise ValueError(
            f"Expected times and signal of shape (n,) but got them of shape {times.shape} and {signal.shape}"
        )

    # preprocess by adding an artificial offset to make
    # the offset fit more stable
    offset = -signal.mean() + (1 + 1j) * 1e3 * signal.std()
    signal = signal + offset

    L = int(L_fraction * len(signal))
    N = len(signal)
    Y = np.zeros((N - L, L), dtype=np.complex128)

    times_diff = np.diff(times)
    if np.max(times_diff - times_diff[0]) > tol:
        raise ValueError("Non uniform sampling times are not supported.")

    # denoise the data using a SVD
    for i in range(L):
        Y[:, i] = signal[i : i + N - L]
    U, S, Vh = np.linalg.svd(Y, False)

    cutoff_idx = n_modes + 1  # set one mode for the constant term

    # filter all eigenvalues lower than the cutoff
    cutoff = np.sort(S)[-cutoff_idx]
    S[S < cutoff] = 0
    Y_filtered = U @ np.diag(S) @ Vh

    # retrieve the filtered signal
    Y1 = Y_filtered[:, :-1]
    Y2 = Y_filtered[:, 1:]

    # compute the eigenvalues of the pencil to find the modes
    Y1_inv = np.linalg.pinv(Y1)
    eigenvalues = np.linalg.eigvals(Y1_inv @ Y2)

    amplitudes, pulsations, decay_rates = _fit_amplitudes(eigenvalues, times, signal, cutoff_idx)
    constant_mode_idx = np.argmin(abs(np.exp(1j * pulsations - decay_rates) - 1))

    offset = amplitudes[constant_mode_idx] - offset
    amplitudes = np.delete(amplitudes, constant_mode_idx)
    pulsations = np.delete(pulsations, constant_mode_idx)
    decay_rates = np.delete(decay_rates, constant_mode_idx)

    return offset, amplitudes, pulsations, decay_rates


def _fit_amplitudes(
    eigenvalues: np.ndarray[complex],
    times: np.ndarray[float],
    signal: np.ndarray[complex],
    n_modes: int,
) -> Tuple[np.ndarray[complex], np.ndarray[complex], np.ndarray[complex]]:
    N = len(times)

    # Vandermonde Matrix
    V = np.zeros((N, n_modes), dtype=eigenvalues.dtype)
    eigenvalues = eigenvalues[np.argsort(abs(eigenvalues))][
        ::-1
    ]  # sort in reversed order
    eigenvalues = eigenvalues[:n_modes]
    acc = np.ones_like(eigenvalues)
    for i in range(N):
        V[i] = acc
        acc = acc * eigenvalues

    coefficients = np.linalg.lstsq(V, signal, rcond=None)[0]

    dt = np.diff(times)[0]
    pulsations = np.angle(eigenvalues) / dt
    decay_rates = -np.log(np.abs(eigenvalues)) / dt

    return coefficients, pulsations, decay_rates


def _match_real_modes(
    amplitudes: np.ndarray[complex],
    pulsations: np.ndarray[float],
    decay_rates: np.ndarray[float],
    tol: float,
) -> Tuple[np.ndarray[float], np.ndarray[float], np.ndarray[float], np.ndarray[float]]:
    """ Associates the complex eigenfrequencies of the fit together to form cosines and sines """

    n = len(amplitudes)
    assert len(amplitudes) == len(pulsations) == len(decay_rates)
    assert n % 2 == 0, "Expected an even number of modes to match real modes"

    normalized_frequency = 1j * abs(pulsations) + decay_rates

    # step 1: separate frequencies that are in the upper and lower complex plane
    positive_w_indices, negative_w_indices = np.where(pulsations > 0)[0], np.where(pulsations < 0)[0]

    # step 2: get their normalized frequencies, i.e. mapped to the upper plane
    positive_w_normalized_frequencies = normalized_frequency[positive_w_indices]
    negative_w_normalized_frequencies = normalized_frequency[negative_w_indices]

    # step 3: create the cost matrix, that is the distance between the normalized frequencies
    positive_w_normalized_frequencies_mg, negative_w_normalized_frequencies_mg = (
        np.meshgrid(
            positive_w_normalized_frequencies, negative_w_normalized_frequencies
        )
    )
    cost = np.abs(
        positive_w_normalized_frequencies_mg - negative_w_normalized_frequencies_mg
    )

    # step 4: use the linear sum assignment to find the best matching pairs
    # and store the initial indices (ie of amplitudes, pulsations, decay_rates) in idx_1 and idx_2
    row_indices, col_indices = linear_sum_assignment(cost)
    idx_1, idx_2 = positive_w_indices[row_indices], negative_w_indices[col_indices]

    # step 5: check if the pairs are valid, i.e. they have close frequencies, decay rates and amplitudes
    if np.any(abs(pulsations[idx_1] - (-pulsations[idx_2])) > tol):
        raise RuntimeError(
            f"All real modes frequencies are expected to have close frequencies "
            f"(within {tol}) but got two paired modes with pulsations {pulsations[idx_1]} and {pulsations[idx_2]}"
        )

    if np.any(abs(decay_rates[idx_1] - decay_rates[idx_2]) > tol):
        raise RuntimeError(
            f"All real modes frequencies are expected to have close decay rates "
            f"(within {tol}) but got two paired modes with decay rates {decay_rates[idx_1]} and {decay_rates[idx_2]}"
        )

    if np.any(abs(amplitudes[idx_1] - np.conj(amplitudes[idx_2])) > tol):
        raise RuntimeError(
            f"All real modes frequencies are expected to have conjugate complex amplitudes "
            f"(within {tol}) but got two paired modes "
            f"with amplitudes {amplitudes[idx_1]} and {amplitudes[idx_2]}"
        )

    # step 6: create the real modes from the matched pairs
    real_pulsations = np.mean([pulsations[idx_1], -pulsations[idx_2]], axis=0).real
    real_decay_rates = np.mean([decay_rates[idx_1], decay_rates[idx_2]], axis=0).real
    real_amplitudes = np.abs([amplitudes[idx_1], amplitudes[idx_2]]).sum(axis=0).real
    real_phases = np.mean(
        [np.angle(amplitudes[idx_1]), -np.angle(amplitudes[idx_2])], axis=0
    ).real

    return real_amplitudes, real_phases, real_pulsations, real_decay_rates



import numpy as np

from .types import FloatLike

"""
    Models are functions that describe the behavior of a system over time.
    They are used both in data fitting and in the generation of synthetic data for the plots.
    In post-processing, we use "adapter" functions to map the array of parameters to the functions arguments.
"""


def _complex_exponential_model(
    t: float | np.ndarray[float],
    offset: complex,
    amplitudes: np.ndarray[complex],
    pulsations: np.ndarray[float],
    decay_rates: np.ndarray[float],
) -> np.ndarray[complex]:
    t = np.array(t)
    return offset + np.sum(
        amplitudes[:, None] * np.exp((1j * pulsations[:, None] - decay_rates[:, None]) * t),
        axis=0,
    ).reshape(t.shape)


def _exponential_model(
    t: float | np.ndarray[float],
    offset: complex,
    amplitudes: np.ndarray[FloatLike],
    decay_rates: np.ndarray[float],
) -> np.ndarray[FloatLike]:
    t = np.array(t)
    return offset + np.sum(
        amplitudes[:, None] * np.exp(-decay_rates[:, None] * t), axis=0
    ).reshape(t.shape)


def _damped_cosine_model(
    t: float | np.ndarray[float],
    offset: float,
    amplitudes: np.ndarray[float],
    phases: np.ndarray[float],
    pulsations: np.ndarray[float],
    decay_rates: np.ndarray[float],
) -> np.ndarray[float]:
    t = np.array(t)
    return offset + np.sum(
        amplitudes[:, None]
        * np.cos(phases[:, None] + pulsations[:, None] * t)
        * np.exp(-decay_rates[:, None] * t),
        axis=0,
    ).reshape(t.shape)
